import { Action } from 'dojo-actions/createAction';
import { ComposeFactory } from 'dojo-compose/compose';
import { EventedListener } from 'dojo-compose/mixins/createEvented';
import { ObservableState, State } from 'dojo-compose/mixins/createStateful';
import { Handle } from 'dojo-core/interfaces';
import Promise from 'dojo-shim/Promise';
import { Child } from 'dojo-widgets/mixins/interfaces';
import { ToAbsMid } from './lib/moduleResolver';
import RegistryProvider from './lib/RegistryProvider';
export { RegistryProvider, ToAbsMid };
/**
 * Any kind of action.
 */
export declare type ActionLike = Action<any, any, any>;
/**
 * Any kind of store.
 */
export declare type StoreLike = ObservableState<State> & {
    add<T>(item: T, options?: any): Promise<T>;
    get<T>(id: string): Promise<T>;
};
/**
 * Any kind of widget.
 */
export declare type WidgetLike = Child;
/**
 * Factory method to (asynchronously) create an action.
 *
 * @param registry The combined registries of the app
 * @param store The store that was defined for this action. It's the factories responsibility to create an action that
 *   observes the store
 * @return The action, or a promise for it
 */
export interface ActionFactory {
    (registry: CombinedRegistry, store: StoreLike): ActionLike | Promise<ActionLike>;
}
/**
 * Factory method to (asynchronously) create a store.
 *
 * @return The store, or a promise for it
 */
export interface StoreFactory {
    (options?: Object): StoreLike | Promise<StoreLike>;
}
/**
 * Factory method to (asynchronously) create a widget.
 *
 * @return The widget, or a promise for it
 */
export interface WidgetFactory {
    (options?: Object): WidgetLike | Promise<WidgetLike>;
}
/**
 * Plain old JavaScript object that contains definitions of actions, stores and widgets.
 */
export interface Definitions {
    /**
     * Action definitions.
     */
    actions?: ActionDefinition[];
    /**
     * Custom element definitions.
     */
    customElements?: CustomElementDefinition[];
    /**
     * Store definitions.
     */
    stores?: StoreDefinition[];
    /**
     * Widget definitions.
     */
    widgets?: WidgetDefinition[];
}
/**
 * Actions, stores and widgets should have string identifiers.
 */
export declare type Identifier = string;
/**
 * Base definition for a single action, store or widget.
 */
export interface ItemDefinition<Factory, Instance> {
    /**
     * Identifier for which the action, store or widget is to be registered.
     */
    id: Identifier;
    /**
     * Factory to create an action, store or widget, or a module identifier that resolves to
     * such a factory.
     */
    factory?: Factory | string;
    /**
     * An action, store or widget instance, or a module identifier that resolves to such
     * an instance.
     */
    instance?: Instance | string;
}
/**
 * Definition for a single action.
 */
export interface ActionDefinition extends ItemDefinition<ActionFactory, ActionLike> {
    /**
     * Identifier of a store which the action should observe for its state.
     *
     * When the action is created it'll automatically observe this store.
     *
     * Note that the `DEFAULT_ACTION_STORE` and `DEFAULT_WIDGET_STORE` identifiers are not supported. The default action
     * store is automatically used if stateFrom is not provided.
     */
    stateFrom?: Identifier | StoreLike;
}
/**
 * Definition for a custom element;
 */
export interface CustomElementDefinition {
    /**
     * The name of the custom element. Must be valid according to
     * <https://www.w3.org/TR/custom-elements/#valid-custom-element-name>.
     */
    name: string;
    /**
     * Factory to create a widget, or a module identifier that resolves to such a factory.
     */
    factory: WidgetFactory | string;
}
/**
 * Definition for a single store.
 */
export interface StoreDefinition extends ItemDefinition<StoreFactory, StoreLike> {
    /**
     * Optional options object passed to the store factory.
     */
    options?: Object;
}
/**
 * Definition for a single widget.
 */
export interface WidgetDefinition extends ItemDefinition<WidgetFactory, WidgetLike> {
    /**
     * Any listeners that should automatically be attached to the widget.
     */
    listeners?: WidgetListenersMap;
    /**
     * Initial state, to be added to the widget's store, if any.
     */
    state?: any;
    /**
     * Identifier of a store which the widget should observe for its state.
     *
     * When the widget is created, the store is passed as the `stateFrom` option.
     *
     * Note that the `DEFAULT_ACTION_STORE` and `DEFAULT_WIDGET_STORE` identifiers are not supported. The default widget
     * store is automatically used if stateFrom is not provided.
     */
    stateFrom?: Identifier | StoreLike;
    /**
     * Optional options object passed to the widget factory. Must not contain `id`, `listeners` and `stateFrom`
     * properties.
     */
    options?: Object;
}
/**
 * A listener for widgets, as used in definitions. May be an identifier for an action or an actual event listener.
 */
export declare type WidgetListener = Identifier | EventedListener<any>;
export declare type WidgetListenerOrArray = WidgetListener | WidgetListener[];
/**
 * A map of listeners where the key is the event type.
 */
export interface WidgetListenersMap {
    [eventType: string]: WidgetListenerOrArray;
}
/**
 * Read-only interface for the combined registries of the app factory.
 */
export interface CombinedRegistry {
    /**
     * Get the action with the given identifier.
     *
     * Note that the action may still need to be loaded when this method is called.
     *
     * @param id Identifier for the action
     * @return A promise for when the action has been loaded. Rejected if loading fails or if no action is registered
     *   with the given identifier.
     */
    getAction(id: Identifier): Promise<ActionLike>;
    /**
     * Check whether an action has been registered with the given identifier.
     *
     * @param id Identifier for the action
     * @return `true` if an action has been registered, `false` otherwise.
     */
    hasAction(id: Identifier): boolean;
    /**
     * Look up the identifier for which the given action has been registered.
     *
     * Throws if the value hasn't been registered.
     *
     * @param action The action
     * @return The identifier
     */
    identifyAction(action: ActionLike): Identifier;
    /**
     * Get the factory for the custom element with the given name.
     *
     * @param name Name of the custom element
     * @return A factory to create a widget for the custom element.
     */
    getCustomElementFactory(name: string): WidgetFactory;
    /**
     * Check whether a custom element has been registered with the given name.
     *
     * @param name Name of the custom element
     * @return `true` if a custom element has been registered, `false` otherwise.
     */
    hasCustomElementFactory(name: string): boolean;
    /**
     * Get the store with the given identifier.
     *
     * Note that the store may still need to be loaded when this method is called.
     *
     * @param id Identifier for the store
     * @return A promise for when the store has been loaded. Rejected if loading fails or if no store is registered
     *   with the given identifier.
     */
    getStore(id: Identifier | symbol): Promise<StoreLike>;
    /**
     * Check whether a store has been registered with the given identifier.
     *
     * @param id Identifier for the store
     * @return `true` if a store has been registered, `false` otherwise.
     */
    hasStore(id: Identifier | symbol): boolean;
    /**
     * Look up the identifier for which the given store has been registered.
     *
     * Throws if the value hasn't been registered.
     *
     * @param store The store
     * @return The identifier
     */
    identifyStore(store: StoreLike): Identifier | symbol;
    /**
     * Create a new widget and add it to the registry.
     *
     * @param factory Factory to create the widget
     * @param options Options to be passed to the factory. Automatically extended with the `registryProvider` option,
     *   and the `stateFrom` option if an `id` was present and the application factory has a default store.
     * @return A promise for a tuple containing the ID of the created widget, and the widget instance itself.
     */
    createWidget<U extends Child, O>(factory: ComposeFactory<U, O>, options?: O): Promise<[string, U]>;
    /**
     * Get the widget with the given identifier.
     *
     * Note that the widget may still need to be loaded when this method is called.
     *
     * @param id Identifier for the widget
     * @return A promise for when the widget has been loaded. Rejected if loading fails or if no widget is registered
     *   with the given identifier.
     */
    getWidget(id: Identifier): Promise<WidgetLike>;
    /**
     * Check whether a widget has been registered with the given identifier.
     *
     * @param id Identifier for the widget
     * @return `true` if a widget has been registered, `false` otherwise.
     */
    hasWidget(id: Identifier): boolean;
    /**
     * Look up the identifier for which the given widget has been registered.
     *
     * Throws if the value hasn't been registered.
     *
     * @param widget The widget
     * @return The identifier
     */
    identifyWidget(widget: WidgetLike): Identifier;
}
export interface AppMixin {
    /**
     * A default store to be used as the `stateFrom` option to action factories, unless another store is specified.
     */
    defaultActionStore?: StoreLike;
    /**
     * A default store to be used as the `stateFrom` option to widget and custom element factories, unless another
     * store is specified.
     */
    defaultWidgetStore?: StoreLike;
    /**
     * Provides access to read-only registries for actions, stores and widgets.
     */
    registryProvider: RegistryProvider;
    /**
     * Register an action with the app.
     *
     * @param id How the action is identified
     * @param action The action to be registered
     * @return A handle to deregister the action
     */
    registerAction(id: Identifier, action: ActionLike): Handle;
    /**
     * Register an action factory with the app.
     *
     * The factory will be called the first time the action is needed. It'll be called with *one* argument:
     * the combined registries of the app.
     *
     * Note that the `createAction()` factory from `dojo-actions` cannot be used here since it requires you to define
     * the `do()` implementation, which the app factory does not allow.
     *
     * @param id How the action is identified
     * @param factory A factory function that (asynchronously) creates an action.
     * @return A handle to deregister the action factory, or the action itself once it's been created
     */
    registerActionFactory(id: Identifier, factory: ActionFactory): Handle;
    /**
     * Register a widget factory for a custom element.
     *
     * The factory will be called each time a widget instance is needed. It may be called with an options argument
     * derived from a `data-options` attribute on the custom element that is to be be replaced by the created widget.
     *
     * @param name The name of the custom element. Must be valid according to
     *   <https://www.w3.org/TR/custom-elements/#valid-custom-element-name>.
     * @param factory A factory that (asynchronously) creates a widget.
     * @return A handle to deregister the custom element
     */
    registerCustomElementFactory(name: string, factory: WidgetFactory): Handle;
    /**
     * Register a store with the app.
     *
     * @param id How the store is identified
     * @param store The store to be registered
     * @return A handle to deregister the store
     */
    registerStore(id: Identifier, store: StoreLike): Handle;
    /**
     * Register a store factory with the app.
     *
     * The factory will be called the first time the store is needed. It'll be called *without* any arguments.
     *
     * @param id How the store is identified
     * @param factory A factory function that (asynchronously) creates a store.
     * @return A handle to deregister the store factory, or the store itself once it's been created
     */
    registerStoreFactory(id: Identifier, factory: StoreFactory): Handle;
    /**
     * Register a widget with the app.
     *
     * @param id How the widget is identified
     * @param widget The widget to be registered
     * @return A handle to deregister the widget
     */
    registerWidget(id: Identifier, widget: WidgetLike): Handle;
    /**
     * Register a widget factory with the app.
     *
     * The factory will be called the first time the widget is needed. It'll be called with an options object
     * that has its `id` property set to the widget ID.
     *
     * @param id How the widget is identified
     * @param factory A factory function that (asynchronously) creates a widget.
     * @return A handle to deregister the widget factory, or the widget itself once it's been created
     */
    registerWidgetFactory(id: Identifier, factory: WidgetFactory): Handle;
    /**
     * Load a POJO definition containing actions, stores and widgets that need to be registered.
     *
     * Action factories will be called with one argument: the combined registries of the app.
     * Store and widget factories will also be called with one argument: an options object.
     *
     * @return A handle to deregister *all* actions, stores and widgets that were registered.
     */
    loadDefinition(definitions: Definitions): Handle;
    /**
     * Take a root element and replace <app-widget> elements with widget instances.
     *
     * @param root The root element that is searched for <app-widget> elements
     * @return A handle to detach rendered widgets from the DOM and remove them from the widget registry
     */
    realize(root: Element): Promise<Handle>;
}
export declare type App = AppMixin & CombinedRegistry;
export interface AppOptions {
    /**
     * A default store to be used as the `stateFrom` option to action factories, unless another store is specified.
     */
    defaultActionStore?: StoreLike;
    /**
     * A default store to be used as the `stateFrom` option to widget and custom element factories, unless another
     * store is specified.
     */
    defaultWidgetStore?: StoreLike;
    /**
     * Function that maps a (relative) module identifier to an absolute one. Used to resolve relative module
     * identifiers in definitions.
     */
    toAbsMid?: ToAbsMid;
}
export interface AppFactory extends ComposeFactory<App, AppOptions> {
}
/**
 * Identifier for the default action store, if any.
 */
export declare const DEFAULT_ACTION_STORE: symbol;
/**
 * Identifier for the default widget store, if any.
 */
export declare const DEFAULT_WIDGET_STORE: symbol;
declare const createApp: AppFactory;
export default createApp;
