import { ComposeFactory } from 'dojo-compose/compose';
import Promise from 'dojo-shim/Promise';
import { Child } from 'dojo-widgets/mixins/interfaces';
import { ActionLike, CombinedRegistry, Identifier, StoreLike, WidgetLike } from '../createApp';
/**
 * Registry to (asynchronously) get instances by their ID.
 */
export interface Registry<I, T> {
    /**
     * Asynchronously get an instance by its ID.
     *
     * @param id Identifier for the instance that is to be retrieved
     * @return A promise for the instance. The promise rejects if no instance was found.
     */
    get(id: I): Promise<T>;
    /**
     * Look up the identifier for which the given value has been registered.
     *
     * Throws if the value hasn't been registered.
     *
     * @param value The value
     * @return The identifier
     */
    identify(value: T): I;
}
/**
 * Registry to (asynchronously) get widget instances by their ID, as well as create new instances that are then added
 * to the registry.
 */
export interface WidgetRegistry<I, T extends Child> extends Registry<I, T> {
    /**
     * Create a new instance and add it to the registry.
     *
     * @param factory Factory to create the new instance
     * @param options Options to be passed to the factory. Automatically extended with the `registryProvider` option,
     *   and the `stateFrom` option if an `id` was present and the application factory has a default store.
     * @return A promise for a tuple containing the ID of the created widget, and the widget instance itself.
     */
    create<U extends T, O>(factory: ComposeFactory<U, O>, options?: O): Promise<[string, U]>;
}
/**
 * Provides access to read-only registries for actions, stores and widgets.
 */
export default class RegistryProvider {
    private actionRegistry;
    private storeRegistry;
    private widgetRegistry;
    private combinedRegistry;
    constructor(combinedRegistry: CombinedRegistry);
    /**
     * Get an action, store or widget registry.
     *
     * @param type The type of registry that is required.
     * @return The registry.
     */
    get(type: 'actions'): Registry<Identifier, ActionLike>;
    get(type: 'stores'): Registry<Identifier | symbol, StoreLike>;
    get(type: 'widgets'): WidgetRegistry<Identifier, WidgetLike>;
    get(type: string): Registry<any, any>;
}
