(function (factory) {
    if (typeof module === 'object' && typeof module.exports === 'object') {
        var v = factory(require, exports); if (v !== undefined) module.exports = v;
    }
    else if (typeof define === 'function' && define.amd) {
        define(["require", "exports", 'dojo-core/lang', 'dojo-shim/Promise', './resolveListenersMap'], factory);
    }
})(function (require, exports) {
    "use strict";
    var lang_1 = require('dojo-core/lang');
    var Promise_1 = require('dojo-shim/Promise');
    var resolveListenersMap_1 = require('./resolveListenersMap');
    function resolveStore(registry, definition) {
        var stateFrom = definition.stateFrom;
        if (!stateFrom) {
            return null;
        }
        if (typeof stateFrom !== 'string') {
            return stateFrom;
        }
        return registry.getStore(stateFrom);
    }
    var errorStrings = {
        action: 'an action',
        customElement: 'a widget',
        store: 'a store',
        widget: 'a widget'
    };
    function isInstance(value) {
        return value && typeof value === 'object';
    }
    function resolveFactory(type, definition, resolveMid) {
        var factory = definition.factory;
        var _a = definition.instance, instance = _a === void 0 ? null : _a;
        if (typeof factory === 'function') {
            return Promise_1.default.resolve(factory);
        }
        if (isInstance(instance)) {
            // <any> hammer since TypeScript can't resolve match the correct overloaded Instance type with the correct
            // Factory return value.
            var factory_1 = function () { return instance; };
            return Promise_1.default.resolve(factory_1);
        }
        var mid = (factory || instance);
        return resolveMid(mid).then(function (defaultExport) {
            if (factory) {
                if (typeof defaultExport !== 'function') {
                    throw new Error("Could not resolve '" + mid + "' to " + errorStrings[type] + " factory function");
                }
                var factory_2 = defaultExport;
                return factory_2;
            }
            // istanbul ignore else Action factories are expected to guard against definitions with neither
            // factory or instance properties.
            if (instance) {
                if (!defaultExport || typeof defaultExport !== 'object') {
                    throw new Error("Could not resolve '" + mid + "' to " + errorStrings[type] + " instance");
                }
                var instance_1 = defaultExport;
                return function () { return instance_1; };
            }
        });
    }
    function makeActionFactory(definition, resolveMid) {
        if (!('factory' in definition || 'instance' in definition)) {
            throw new TypeError('Action definitions must specify either the factory or instance option');
        }
        if ('instance' in definition && 'stateFrom' in definition) {
            throw new TypeError('Cannot specify stateFrom option when action definition points directly at an instance');
        }
        return function (registry, defaultActionStore) {
            return Promise_1.default.all([
                resolveFactory('action', definition, resolveMid),
                resolveStore(registry, definition)
            ]).then(function (_a) {
                var _factory = _a[0], _store = _a[1];
                var factory = _factory;
                var store = _store || defaultActionStore;
                return factory(registry, store);
            });
        };
    }
    exports.makeActionFactory = makeActionFactory;
    function makeCustomElementFactory(definition, resolveMid) {
        var promise;
        var factory;
        return function (options) {
            if (factory) {
                return factory(options);
            }
            if (!promise) {
                // Memoize the factory resolution.
                promise = resolveFactory('customElement', definition, resolveMid).then(function (result) {
                    factory = result;
                    promise = null;
                });
            }
            return promise.then(function () {
                return factory(options);
            });
        };
    }
    exports.makeCustomElementFactory = makeCustomElementFactory;
    function makeStoreFactory(definition, resolveMid) {
        if (!('factory' in definition || 'instance' in definition)) {
            throw new TypeError('Store definitions must specify either the factory or instance option');
        }
        if ('instance' in definition && 'options' in definition) {
            throw new TypeError('Cannot specify options when store definition points directly at an instance');
        }
        var options = lang_1.assign({}, definition.options);
        return function () {
            return resolveFactory('store', definition, resolveMid).then(function (factory) {
                return factory(options);
            });
        };
    }
    exports.makeStoreFactory = makeStoreFactory;
    function makeWidgetFactory(definition, resolveMid, registry) {
        if (!('factory' in definition || 'instance' in definition)) {
            throw new TypeError('Widget definitions must specify either the factory or instance option');
        }
        if ('instance' in definition) {
            if ('listeners' in definition) {
                throw new TypeError('Cannot specify listeners option when widget definition points directly at an instance');
            }
            if ('state' in definition) {
                throw new TypeError('Cannot specify state option when widget definition points directly at an instance');
            }
            if ('stateFrom' in definition) {
                throw new TypeError('Cannot specify stateFrom option when widget definition points directly at an instance');
            }
            if ('options' in definition) {
                throw new TypeError('Cannot specify options when widget definition points directly at an instance');
            }
        }
        var rawOptions = definition.options;
        if (rawOptions) {
            if ('id' in rawOptions || 'listeners' in rawOptions || 'stateFrom' in rawOptions) {
                throw new TypeError('id, listeners and stateFrom options should be in the widget definition itself, not its options value');
            }
            if ('registryProvider' in rawOptions) {
                throw new TypeError('registryProvider option must not be specified');
            }
        }
        return function (_a) {
            var registryProvider = _a.registryProvider, defaultWidgetStore = _a.stateFrom;
            var id = definition.id, initialState = definition.state;
            var options = lang_1.assign({
                id: id,
                registryProvider: registryProvider
            }, rawOptions);
            return Promise_1.default.all([
                resolveFactory('widget', definition, resolveMid),
                resolveListenersMap_1.default(registry, definition.listeners),
                resolveStore(registry, definition)
            ]).then(function (_a) {
                var _factory = _a[0], _listeners = _a[1], _store = _a[2];
                var factory = _factory;
                var listeners = _listeners;
                var store = _store || defaultWidgetStore;
                if (listeners) {
                    options.listeners = listeners;
                }
                if (store) {
                    options.stateFrom = store;
                }
                if (store && initialState) {
                    return store.add(initialState, { id: id })
                        .catch(function () { return undefined; })
                        .then(function () { return factory(options); });
                }
                return factory(options);
            });
        };
    }
    exports.makeWidgetFactory = makeWidgetFactory;
});
//# sourceMappingURL=factories.js.map